
# Copyright (C)  2004 by Intevation GmbH
# Authors:
# Ludwig Reiter <ludwig@intevation.de>
# Silke Reimer <silke@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

import sys
import edbsclasses
import logging
from string import join
log = logging.getLogger('')

def delete(ONS, sqlfile):
    """Diese Funktion geht die ONS durch und sammlet alle Objnr mit
    Lschkommandos. Dann werden die DELETE-Kommados fr die einzelnen
    Tabellen geschrieben"""

    #SQL_Kom = ""
    # Die Liste der Objektnummern, der zu lschendenden Eintrge 
    zuloeschenObje = []

    # Durchgehen der Objnr im ONS
    for Objnr in ONS.keys():
        # Ist der Eintrag passend 
        if not ONS[Objnr].compareObjnr(Objnr):
            print Objnr , ONS[Objnr].compareObjnr(Objnr)  
            raise "error: Ein objdaten_sammler hat andere Objnr als im ONS steht" 

        # Falls das Objekt zu lschen ist, fge es in die zuloeschenObje
        # -Liste ein
        if ONS[Objnr].Status in ['L','M']:
            zuloeschenObje.append(Objnr)

    # Erzeugen der SQL-Kommandos
    if zuloeschenObje:
        # Geht die Objnr der zu loschenden Eintrge durch
        # und Teilt sie in 225 Gruppen.
        SQL_Objnrn_liste =[]
        while zuloeschenObje:
            jetzt_loeschen = ["'%s'" % obj for obj in zuloeschenObje[:220]]
            zuloeschenObje = zuloeschenObje[220:]
            SQL_Objnrn_liste.append(', '.join(jetzt_loeschen))

        # Geht die Tabellen durch und schreibt fr jeden SQL_Objnrn-String
        # einen SQL-Befehl
        for tab in edbsclasses.Tabnamen:
            if tab:
                for SQL_Objnrn in SQL_Objnrn_liste:
                    sqlfile.write("DELETE FROM %s WHERE objnr IN (%s) ; \n"
                    % (tab, SQL_Objnrn) )

def exportiere_SQL(ONS , sqlfile):
    """Diese Funktion erhlt einen Objektnummernspeicher (ONS)(, der zum
    Beispiel mit second_parsing erzeugt wurde). Zunchste werden alle
    Objekte mit Lsch-bit gelscht, danach werden alle (Neu-)Eintrge in
    die Datenbank abgearbeitet. Die erzeugten SQL-Befehle werden in die
    Datei 'Ziel_SQL' geschrieben.
    """

    # Zuerst werden die Lschoperationen bearbeitet
    delete(ONS, sqlfile)

    # Fr alle Objnr die bekannt, werden die INSERT-Befehle erezugt
    print "Trage %d Objekte ein" % len(ONS)
    for Objnr in ONS.keys():
        if not ONS[Objnr].compareObjnr(Objnr):
            print Objnr , ONS[Objnr].compareObjnr(Objnr)  
            raise "error: Ein objdaten_sammler hat andere Objnr als im ONS steht"

        # Danach werden die INSERT-operationen bearbeitet
        #SQL_Daten = ONS[Objnr].export_Oracle()
        SQL_Daten = export(ONS[Objnr])
        sqlfile.write(SQL_Daten)

    # Wegschreiben der Logeintrge
    log.warning(edbsclasses.Punktfehler)
    edbsclasses.Punktfehler = ""

def anpassen_Index(Tabnamen, sqlfile):
    """Kriegt eine Liste von Tabellennamen bergeben und erstellt die
    ALTER_INDEX__REBUILD-SQL-Kommandos fr die Tabellennamen, die mit
    _p, _l, _f enden. Der angepasste Index ist der <Tabellenname>_idx."""

    # Erzeuge die Alter Index Rebuild -SQL-Kommandos jetzt hier
    for tabname in Tabnamen:
        if tabname[-2:] in ['_p', '_l', '_f' ]:
            sqlfile.write("ALTER INDEX %s_idx REBUILD; \n" % tabname)


def get_SQL_geoangaben(objektteil):
    """Hier wird aus den Geometrieangaben SQL-Befehle gemacht. Es gibt
       die Mglichkeit fr PostgreSQL(PG) und Oracle DB die passenden
       SQL_Befehle zu schreiben ber die Variable format. Geht
       die artdergeo durch(ob 11,15,51,52,53) und gibt die passenden
       SQL_Befehle zurck.
       """
    SQL_Kom = ""
    winkel = 0

    # Linie
    if objektteil.artderGeo in ['11', '15']:
        linie = objektteil.geoangaben()
        linelist = []
        coord_anzahl = 0 # nach 30 coord-Paare ein newline einfgen
        # Alle Coordinaten eintragen...
        for (x,y) in linie:
            if coord_anzahl == 30:
                endline = '\n'
                coord_anzahl = 0
            else:
                endline = ''
            linelist.append('%s,%s%s' %(round(x,3),round(y,3), endline))
            coord_anzahl += 1
        #... und zusammenfgen
        linienstr = join(linelist,',')

        declareSQL = "\
DECLARE\ngeom mdsys.sdo_geometry :=\n\
MDSYS.SDO_GEOMETRY(\
2002, null, null, MDSYS.SDO_ELEM_INFO_ARRAY(1, 2, 1), \
MDSYS.SDO_ORDINATE_ARRAY(%s));\n" % linienstr

    # Punkt
    else:
        # ohne Winkelangabe
        if objektteil.artderGeo in ('51', '52'):
            (punkt_x, punkt_y) = objektteil.geoangaben()
        # mit Winkelangabe
        elif objektteil.artderGeo == '53':
            (winkel,(punkt_x, punkt_y)) = objektteil.geoangaben()
        declareSQL = "\
DECLARE\ngeom mdsys.sdo_geometry :=\n\
MDSYS.SDO_GEOMETRY(2001, null, \
MDSYS.SDO_POINT_TYPE(%s,%s,null), null, null);\n"\
            % (str(round(punkt_x,3)), str(round(punkt_y,3)))

    SQL_Kom += declareSQL
    return (SQL_Kom,winkel)

def anpassen(Text):
    """Passt den Text so an, dass in Text ' richtig in die DB kommt.
       Also in einen String Text wir das einfache Anfuehrungszeichen
       ' verdoppelt und sorgt dafr, dass die hinternen Leerzeichen
       verschwinden."""           
    Text_angepasst = ""
    for ch in Text:
        if ch == "'":
            Text_angepasst += "''"
        else:                  
            Text_angepasst += ch
    Text_angepasst= Text_angepasst.rstrip()
    return Text_angepasst

def get_BesondereInfo(obj, objektteil):
    """Diese Methode kriegt die objnr und die Stelle der BI(beta)
      bergeben und gibt ein Paar([DatenfuerFolie],
      String mit SQL-Kommandos, FehlerLog) zurck.
      Die DatenfrFolie haben folgendes Format:
       [ text, kartentyp, artderinfo , darstellungsart]
      Zum Erstellen der SQL_Kommands holt sich diese Methode
      je nach Folie,ArtderInfo und Darstellunginfo(SpeziSch)
      die Vorgehensweise aus dem Spezifikations-Directory Spezi.
      Diese Spezisatz enthlt die Vorgehensweise: in welche Tabelle
      geschrieben
      wird(mglicherweise sollen die Daten zurckgegeben werden, damit sie in 
      die Folientabelle kommen), auf welche Zusatzattributte die Text -Daten
      der Besondenderen Infos verteilt werden. 
      """
    DatenfuerFolie = []  
    SQL_Kom = "" 

    # Folie aus FktdesObj
    Folie = obj.folie
    # Objnr aus FktdesObj
    Objnr = obj.Objnr
    # Art der Besonderen Info 
    ArtderInfo = objektteil.artderInfo
    # Holt die Darstllungsart
    Dar = objektteil.darstellung
    ArtderGeo =objektteil.artderGeo
    # Spezifikations- Schlssel
    SpeziSch = Folie + ArtderInfo + Dar 

    if edbsclasses.Spezi.has_key(SpeziSch):
        if edbsclasses.Spezi[SpeziSch][0] == "" and ArtderGeo == "  ":
            DatenfuerFolie = [ objektteil.artderGeo,
                               objektteil.kartentyp,
                               ArtderInfo,
                               Dar  ]
        elif edbsclasses.Spezi[SpeziSch][0] =="" :
            log.warning("warnung: In Folientabelle wird geschrieben trotz Geometrieangaben Folie/AdBI/Dar: %s/%s/%s\n" %(Folie, ArtderInfo, Dar) )
        else:
            Spezisatz = edbsclasses.Spezi[SpeziSch]

            # Index fr die passenden Zeichen aus BesondereInfo Text
            m=0

            # Schaut die Geodaten durch
            if ArtderGeo in ['51','52','53', '11', '15']:
                (SQL_Kom, winkel) = get_SQL_geoangaben(objektteil)

            # Attribute und AttributListen erstellen
            # Attribute die immer vorkommen
            # fr die verschiedenen Attribute
            SQL_Att = ['objnr', 'kartentyp', 'darstellungsart']
            # fr die Werte der Attribute
            SQL_Val = ["'%s'" % Objnr, \
                "'%s'" % objektteil.kartentyp, \
                '%s' % Dar]

            # Attribute aus der Attributsliste
            for n in range(len(Spezisatz[1]) /2):
                 if m >= 33:
                    log.warning("warung: Zuviele Textzeichen abgefragt fuer\
                    besondere Information von Objekt %s (BI: %s)" \
                        %(Objnr, objektteil.BI))
                    # Schon genug Zeichen gelesen
                    break

                 if Spezisatz[1][2*n] != "":
                     SQL_Att.append(Spezisatz[1][2*n])
                     value = anpassen(objektteil.BI[m:m+Spezisatz[1][2*n+1]])
                     SQL_Val.append("'%s'" % value)

                 m += Spezisatz[1][2*n+1]   

            # Geometry -Attribut
            if ArtderGeo in ['11','15', '51', '52', '53']:
                 SQL_Att.append('geom')
                 SQL_Val.append('geom')
            if ArtderGeo in ['51','52','53']:
                 SQL_Att.append('winkel')
                 SQL_Val.append('%d' % winkel)


            objtyp = ''
            if ArtderGeo in ['51','52','53']:
                objtyp = "_p"
            elif ArtderGeo in ['11','15']:
                objtyp = "_l"
            tabname = edbsclasses.Spezi[SpeziSch][0] + objtyp
            insertsql = "\
BEGIN\n\
INSERT INTO %s \
(%s) VALUES(%s);\n\
COMMIT;\nEND;\n/\n"\
% (tabname, join(SQL_Att, ', '), join(SQL_Val, ', '))

            SQL_Kom += insertsql
    else:
        log.warning("warning: eine besondere Info konnte nicht zugeordnet werden \n" + Objnr)
        log.warning("Folie: " + Folie +" ArtderInfo: "+ ArtderInfo+" Darstellungsart: "+ Dar+"\n")
    return [DatenfuerFolie, SQL_Kom]


def get_SQL_Flaeche(obj):
    """Gibt je nach format die Flaechedaten in SQL-Kommandos zurck."""

    polygonliste = obj.get_Polygon()
    if polygonliste == []:
        print "no values"
        return ""

    index = 1
    coords = []
    elements = []
    for polygon in  polygonliste:
        coordlist = polygon.getganzeListe()
        elements.append("%s,1003,1" %index)
        index += len(coordlist) * 2

        coord_anzahl = 0 # nach 30 coord-Paaren ein newline einfgen
        # Alle Coordinaten eintragen...
        for (x,y) in coordlist:
            if coord_anzahl == 30:
                endline = '\n'
                coord_anzahl = 0
            else:
                endline = ''
            coords.append('%s,%s%s' %(round(x,3),round(y,3), endline))
            coord_anzahl += 1
        for loch in polygon.getLoecher():
            lochcoordlist = loch.getganzeListe()
            if len(lochcoordlist) > 2:
                elements.append("%s,2003,1" %index)
                index += len(lochcoordlist) * 2
                coord_anzahl = 0 # nach 30 coord-Paaren ein newline einfgen
                # Alle Coordinaten eintragen...
                for (x,y) in lochcoordlist:
                    if coord_anzahl == 30:
                        endline = ',\n'
                        coord_anzahl = 0
                    else:
                        endline = ''
                    coords.append('%s,%s%s' %(round(x,3),round(y,3), endline))
                    coord_anzahl += 1
            else: 
                print "Ein Loch in diesem Polygon hat weniger als 3 Koordinaten."

    SQL_Kom = """DECLARE\ngeom mdsys.sdo_geometry :=
MDSYS.SDO_GEOMETRY(2007, null, null, MDSYS.SDO_ELEM_INFO_ARRAY(%s), MDSYS.SDO_ORDINATE_ARRAY(%s));
""" % (join(elements,','), join(coords,','))
    return SQL_Kom


def get_SQL_Linie(obj):
    """Gibt je nach format die Liniedaten in SQL-Kommandos zurck."""

    #Holt sich die Linien
    Linienzugliste = obj.get_ganzeLinie()

    coords = []
    elements = []
    index = 1
    for linie in  Linienzugliste:
        coordlist = linie.getganzeListe()
        elements.append("%s,2,1" %index)
        index += len(coordlist) * 2

        coord_anzahl = 0 # nach 30 coord-Paaren ein newline einfgen
        # Alle Coordinaten eintragen...
        for (x,y) in coordlist:
            if coord_anzahl == 30:
                endline = '\n'
                coord_anzahl = 0
            else:
                endline = ''
            coords.append('%s,%s%s' %(round(x,3),round(y,3), endline))
            coord_anzahl += 1

    SQL_Kom = """DECLARE\ngeom mdsys.sdo_geometry :=
MDSYS.SDO_GEOMETRY(2006, null, null, MDSYS.SDO_ELEM_INFO_ARRAY(%s), MDSYS.SDO_ORDINATE_ARRAY(%s));
""" % (join(elements,','), join(coords,','))
    return SQL_Kom

def get_SQL_Punkt(obj):
    """Gibt je nach format die Punktdaten in SQL-Kommandos zurck."""

    if edbsclasses.format == "Oracle":
        (x, y) = edbsclasses.punkttrans(obj.Objkoor)
        SQL_Kom = """
        DECLARE
        geom mdsys.sdo_geometry :=
        MDSYS.SDO_GEOMETRY( 2001, null, MDSYS.SDO_POINT_TYPE(%s, %s, null),
        null, null);
        """ % (x, y)

    return SQL_Kom

def get_SQL_Folie(obj, BI):
    """geht die Folien durch und gibt jeweils das SQL_Kommando fr
    ein Objekt zurck. Nutzt die Methoden get_SQL_Linie, get_SQL_Punkt
    und get_SQL_Flaeche fr die Geometriedaten"""
    Objart = obj.art # Holt die Objektart
    Objtyp = obj.typ   # Holt den Objekttyp : P,L,F
    #FehlerLog =""

    # Erstellen von Attribut und Werte Tupel 
    SQL_Att = []
    SQL_Val = []
    SQL_Att.append("objnr")
    SQL_Val.append("'%s'" %obj.Objnr)
    SQL_Att.append("objart")
    SQL_Val.append("%s" %Objart)
    SQL_Att.append("aktualitaet")
    SQL_Val.append("'%s'" %obj.aktualitaet)
    SQL_Att.append("edatum")
    SQL_Val.append("to_date('%s', 'RRMMDD')" % obj.datum)
    # Attribute aus den BI
    if BI != []:
        m = 10
        SpeziSch = obj.folie + BI[2] + BI[3]
        SQL_Att.append('kartentyp')
        SQL_Att.append('darstellungsart')
        SQL_Val.append("'%s'" %BI[1])
        SQL_Val.append("%s" %BI[3])
        
        if edbsclasses.Spezi.has_key(SpeziSch):
            Spezisatz =edbsclasses.Spezi[SpeziSch]
            for n in range(len(Spezisatz[1]) /2):
                if m >= 33:
                    log.error("warnung: Obj %s Zuviele Textzeichen werden\
                    abgefragt\n" % obj.Objnr)
                    break
                if Spezisatz[1][2*n] != "":
                    SQL_Att.append(Spezisatz[1][2*n])
                    value = anpassen(BI[0][m:m+Spezisatz[1][2*n+1]])
                    SQL_Val.append("'%s'" %value)
                m += Spezisatz[1][2*n+1]    
        else:
            log.error("Nicht erkannter Spezifikationsschlssel in get_SQL_Folie \n")

    # Erstellen des INSERT_Befehles
    # Erstellen der DECLARE -Kommandos
    if Objtyp =="F":
        SQL_geom = get_SQL_Flaeche(obj)
        if SQL_geom == '':
            print "Objekt %s: Die Linien lassen sich nicht zu einem Polygon zusammenfgen" % obj.Objnr
            return ''
    elif Objtyp == "L":
        SQL_geom = get_SQL_Linie(obj) 
    elif Objtyp == "P":
        SQL_geom = get_SQL_Punkt(obj)

    if not edbsclasses.tables.has_key(obj.folie):
        print "Tabellenspezifikation fr Folie %s ist nicht vorhanden" % obj.folie
        return ''
    tabname = edbsclasses.tables[obj.folie].tabname + '_' + Objtyp.lower()
    SQL_Kom = SQL_geom
    # Der Prozedure Teil des INSERT-Kommandos wird erstellt
    insertsql = "\
BEGIN\n\
    INSERT INTO %s (%s, geom) VALUES(%s, geom);\n\
    COMMIT;\nEND;\n/\n" \
    %(tabname, join(SQL_Att,', '), join(SQL_Val, ', '))

    SQL_Kom += insertsql

    return SQL_Kom

def export(obj):
    """Diese Methode erzeugt die  SQL_Kommados fr diesen
    objdaten_sammler.
    Falls obj.Status in ['N','M'], werden alle BesonderenInfos mit 
    obj.get_BesondereInfos abgearbeitet und das ergebnis zu
    SQL_Kom concated und der SQL_Befehl fr die Folie-Tabelle 
    erstellt und concated"""
    SQL_Kom =""

    if obj.Status in ['M','N']:
        BIfuerFolie = []
        if obj.II_daten != None:
            for ot in obj.objektteile:
                Temp = get_BesondereInfo(obj, ot) 
                BIfuerFolie = Temp[0]
                SQL_Kom_tmp = Temp[1]
                SQL_Kom += SQL_Kom_tmp

            SQL_Kom_tmp = get_SQL_Folie(obj, BIfuerFolie)
            SQL_Kom += SQL_Kom_tmp
        else:
            log.warning("warning: Objekt mit Nummer: " + obj.Objnr +" hat None als obj.II_daten \n")
    return SQL_Kom

