
# Copyright (C)  2004 by Intevation GmbH
# Authors:
# Ludwig Reiter <ludwig@intevation.de>
# Silke Reimer <silke@intevation.de>
#
# This program is free software under the GPL (>=v2)
# Read the file COPYING coming with the software for details.

import sys
import edbsclasses
import logging
import ogr
from string import join
log = logging.getLogger('')


def anpassen(Text):
    """Passt den Text so an, dass in Text ' richtig in die DB kommt.
       Also in einen String Text wir das einfache Anfuehrungszeichen
       ' verdoppelt und sorgt dafr, dass die hinternen Leerzeichen
       verschwinden."""           
    Text_angepasst = ""
    for ch in Text:
        if ch == "'":
            Text_angepasst += "''"
        else:                  
            Text_angepasst += ch
    Text_angepasst= Text_angepasst.rstrip()
    return Text_angepasst


class OGRCollection:
    """ Diese Klasse dient dazu, die nach der EDBS-Philosophie
    gehaltenen Daten aus dem ONS in die Tabellenstruktur aus der
    edbsilon-Konfigurationsdatei zu bringen. Anschlieend knnen
    die Daten ber eine mit OGR ansprechbare Connection in ein
    anderes Datenformat exportiert werden
    """ 

    def __init__(self, tables, datasource='PG:dbname=alktest host=eukrante user=silke', driver="PostgreSQL"):
        self.tables = {}
        for t in tables:
            self.tables[t] = []
        self.datasource = datasource
        self.driver = driver
        self.deletelist = []

    def readONS(self, ONS):
        """Diese Funktion erhlt einen Objektnummernspeicher (ONS)(, der zum
        Beispiel mit second_parsing erzeugt wurde). 
        Zunchste werden alle
        Objekte mit Lsch-bit gelscht, danach werden alle (Neu-)Eintrge in
        die Datenbank abgearbeitet. Die erzeugten SQL-Befehle werden in die
        Datei 'Ziel_SQL' geschrieben.
        """

        # Zuerst werden die Lschoperationen bearbeitet
        #delete(ONS, sqlfile)

        # Fr alle Objnr die bekannt, werden die OGR-Feature erzeugt
        print "Trage %d Objekte ein" % len(ONS)
        for Objnr in ONS.keys():
            if not ONS[Objnr].compareObjnr(Objnr):
                print Objnr , ONS[Objnr].compareObjnr(Objnr)  
                raise "error: Ein objdaten_sammler hat andere Objnr als im ONS steht"
            # Danach werden die INSERT-operationen bearbeitet
            self.readObj(ONS[Objnr])

        # Wegschreiben der Logeintrge
        log.warning(edbsclasses.Punktfehler)
        edbsclasses.Punktfehler = ""

    def readObj(self, obj):
        """Diese Methode erzeugt die  SQL_Kommados fr diesen
        objdaten_sammler.
        Falls obj.Status in ['N','M'], werden alle BesonderenInfos mit 
        obj.get_BesondereInfos abgearbeitet und das ergebnis zu
        SQL_Kom concated und der SQL_Befehl fr die Folie-Tabelle 
        erstellt und concated"""

        if obj.Status in ['M','N']:
            if obj.II_daten != None:
                attributes = {}
                for beta in range(len(obj.II_daten[1])):
                    biattr =  self.insertBI(obj, beta) 
                    for a in biattr.keys():
                        attributes[a] = biattr[a]

                attributes['objnr'] = obj.Objnr
                attributes['objart'] = obj.art
                attributes['aktualitaet'] = obj.aktualitaet
                monate = []
                monate.append('jan')
                monate.append('feb')
                monate.append('mar')
                monate.append('apr')
                monate.append('may')
                monate.append('jun')
                monate.append('jul')
                monate.append('aug')
                monate.append('sep')
                monate.append('oct')
                monate.append('nov')
                monate.append('dec')
                tag = obj.tag
                monat = monate[int(obj.monat) -1]
                jahr = int(obj.jahr)
                if jahr < 30:
                    attributes['edatum'] = '%s.%s.20%s' %(tag, monat, jahr)
                else:
                    attributes['edatum'] = '%s.%s.19%s' %(tag,monat,jahr)
    
                # Geometrie
                if Objtyp =="F":
                    geom = getFlaeche(obj)
                elif Objtyp == "L":
                    geom = getLinie(obj) 
                elif Objtyp == "P":
                    geom = getPunkt(obj)
                if geom == '':
                    print "Objekt %s: Das Objekt hat keine korrekte Geometrie" % obj.Objnr
                else:
                    attributes['geom'] = geom
    
                tabname = "f"+ folie +"_"+ Objtyp.lower()
                if not self.tables.has_key(tabname):
                    print "Eine Folientabelle fr Foliennr %s ist nicht spezifiziert" % folie
                else:
                    self.tables[tabname].append(attributes)

            else:
                print "warning: Objekt %s:  hat None als obj.II_daten" % obj.Objnr
        if obj.Status in ['L','M']:
            self.deletelist.append(obj.Objnr)


    def delete(self, ds):
        """Diese Funktion geht die ONS durch und sammlet alle Objnr mit
        Lschkommandos. Dann werden die DELETE-Kommados fr die einzelnen
        Tabellen geschrieben"""

        if len(self.deletelist) == 0:
            return
        for i in  range(ds.GetLayerCount()):
            layer =  ds.GetLayer(i)
            layer.GetFeatureCount()
            f = layer.GetNextFeature()
            while f:
                f.GetFieldIndex('objnr')
                objnr = f.GetField(f.GetFieldIndex('objnr'))
                if objnr in self.deletelist:
                    layer.DeleteFeature(f.GetFID())
                f = layer.GetNextFeature()


    def export(self):
        """ Exportieren der Elemente"""
        driver = ogr.GetDriverByName(self.driver)
        ds = driver.Open(self.datasource, 1)
        if ds == None:
            ds = driver.CreateDataSource(self.datasource)
        self.delete(ds)
        for t in self.tables.keys():
            featurelist = self.tables[t]
            if len(featurelist) != 0:
                try:
                    layer = ds.GetLayerByName( t )
                except:
                    # es gibt mindestens ein Objekt in der featurelist.
                    # Setze Geometry-Typ des ersten Objektes als geomtype
                    # fest
                    feature = featurelist[0]
                    if feature.has_key('geom'):
                        geomtype = feature['geom'].GetGeometryType()
                        layer = ds.CreateLayer( t, None, geomtype) 
                    else:
                        layer = ds.CreateLayer( t )
                    for a in self.tables[t][0].keys():
                        if a != 'geom':
                            field = ogr.FieldDefn(a, ogr.OFTString)
                            layer.CreateField(field)
                for f in self.tables[t]:
                    feature = ogr.Feature( layer.GetLayerDefn() ) 
                    for a in f.keys():
                        if a != 'geom':
                            feature.SetField(a, f[a])
                        else:
                            feature.SetGeometry(f[a])
                    layer.CreateFeature(feature)
        ds.Destroy()

    def insertBI(self, obj, beta):
        """ Die besonderen Informationen werden als
        Zusatzinformationen zu einem EDBS-Objekt gespeichert. Da
        die besonderen Informationen zustzlich auch
        Geographische Angaben enthalten knnen, wird ein Teil
        u.U. in eigenen Tabellen abgelegt. 
        Dies geschieht in dieser Methode. 
        """

        # Folie aus FktdesObj
        folie = obj.folie()
        # Art der Besonderen Info 
        artderInfo =obj.II_daten[1][beta][0][0][0:2]
        # Holt die Darstllungsart
        darstellung = obj.II_daten[1][beta][0][0][6:10]
        # Spezifikations- Schlssel
        schluessel = folie + artderInfo + darstellung

        artderGeo = obj.II_daten[1][beta][0][0][43:45]

        if edbsclasses.Spezi.has_key(schluessel):
            Spezisatz = edbsclasses.Spezi[schluessel]
            # Attribute aus der Attributsliste
            attributes = {}
            for n in range(len(Spezisatz[1])/2):
                m = 10
                if m >= 43:
                    log.warning("Warnung: Zuviele Textzeichen abgefragt " + Objnr +" "+beta)
                    # Schon genug Zeichen gelesen
                    break  
                if Spezisatz[1][2*n] != "":
                    value = obj.II_daten[1][beta][0][0][m:m+Spezisatz[1][2*n+1]]
                    attributes[Spezisatz[1][2*n]] = value
                    m += Spezisatz[1][2*n+1]   
            if edbsclasses.Spezi[schluessel][0] == "" :
                if artderGeo == "  ":
                    return attributes
                else:
                    print("Warnung: In Folientabelle wird geschrieben trotz Geometrieangaben Folie/AdBI/Dar: %s/%s/%s\n" %(folie, artderInfo, darstellung) )
            else:
                # Objnr aus FktdesObj
                Objnr = obj.Objnr

                # Index fr die passenden Zeichen aus BesondereInfo Text
                # 10 ist der Startindex des Textfeldes 
                m=10  

                # Attribute und AttributListen erstellen
                # Attribute die immer vorkommen
                attributes['objnr'] = Objnr
                attributes['kartentyp'] = obj.II_daten[1][beta][0][0][2:4]
                attributes['darstellungsart'] = darstellung

                # Eintragen der Geometrie-Daten falls vorhanden
                #(artderGeo in (51,52,53) fehlt
                if artderGeo in ['51','52','53', '11', '15']:
                    geom, winkel = getGeometrie(obj, artderGeo, beta)
                    attributes['geom'] = geom
                    if winkel != None:
                        attributes['winkel'] = winkel

                # Tabellenname zusamensetzen und Attributliste
                # dort ablegen:
                type = ''
                if artderGeo in ['51','52','53']:
                    type =  "_p"
                elif artderGeo in ['11','15']:
                    type =  "_l"
                table = edbsclasses.Spezi[schluessel][0]  + type
                if not self.tables.has_key(table):
                    self.tables[table] = []
                self.tables[table].append(attributes)
                return {}
        else:
            print "Eine besondere Info konnte nicht zugeordnet werden" 
            print "    Folie: %s, artderInfo: %s, Darstellungsart: %s"\
                %(folie, artderInfo, darstellung)
            return {}


def getFlaeche(obj):
    """Gibt die Flche eines Objektes als
    OGR-Polygon-Geometry zurck
    """

    polygonliste = obj.get_Polygon()
    if polygonliste == []:
        return ""
    else:
        multipolygon = []
        for polygon_tmp in polygonliste:
            polygonAsList = []
            coordAsList = []
            for (x,y) in polygon_tmp.getganzeListe():
                coordAsList.append('%s %s' %(round(x,3),round(y,3)))
            polygonAsList.append('(%s)' % join(coordAsList,',') )
            loecher = polygon_tmp.getLoecher()
            if len(loecher) == 0:
                pass
            else:
                for loch in polygon_tmp.getLoecher():
                    coordAsList = []
                    for (x,y) in loch.getganzeListe():
                        coordAsList.append('%s %s' %(round(x,3),round(y,3)))
                    polygonAsList.append('(%s)' % join(coordAsList,',') )
            multipolygon.append('(%s)' % join(polygonAsList, ','))

        return ogr.CreateGeometryFromWkt('MULTIPOLYGON(%s)'
            %join(multipolygon, ','))

def getLinie(obj):
    """Gibt die Linie eines Objektes als
    OGR-LINESTRING-Geometrie zurck
    """
    #Holt sich die Linien
    Linienzugliste = obj.get_ganzeLinie()
    multiline = []
    for Linienzug in Linienzugliste:
        coordAsList = []
        for(x,y) in Linienzug.getganzeListe():
            coordAsList.append('%s %s' %(round(x,3),round(y,3)))
        multiline.append('(%s)' %join(coordAsList,','))
    return ogr.CreateGeometryFromWkt('MULTILINESTRING(%s)' %join(multiline))

def getPunkt(obj):
    """Gibt einen Punkt eines Objekes als
    ogr-POINT-Geometrie zurck
    """
    (x, y) = edbsclasses.punkttrans(obj.Objkoor)
    return ogr.CreateGeometryFromWkt('POINT(%s %s)' %(x,y))

def getGeometrie(obj, artdergeo,beta):
    """Hier wird aus den Geometrieangaben SQL-Befehle gemacht. Es gibt
       die Mglichkeit fr PostgreSQL(PG) und Oracle DB die passenden
       SQL_Befehle zu schreiben ber die Variable format. Geht
       die artdergeo durch(ob 11,15,51,52,53) und gibt die passenden
       SQL_Befehle zurck.
       """

    # Linie
    if artdergeo in ['11', '15']:

        linie = obj.get_geoangaben(int(artdergeo),beta)
        Komma = 0
        zeilenzahl = 0 # Damit die Zeilen nicht zu lang werden
                       # wird mglicherweise ein "\n" eingefgt 
        linienstr = []
        for (x,y) in linie:
            linienstr.append('%s %s' %(round(x,3),round(y,3)))

        return ogr.CreateGeometryFromWkt('LINESTRING(%s)' %join(linienstr,',')), None

    # Punkt ohne Winkelangabe
    elif artdergeo == '51' or artdergeo == '52':
        (x,y) = obj.get_geoangaben(51,beta)
        return ogr.CreateGeometryFromWkt('POINT(%s %s)' %(x,y)), None

    # Punkt mit Winkelangabe
    elif artdergeo == '53':
        (winkel,(x, y)) = obj.get_geoangaben(53,beta)
        return ogr.CreateGeometryFromWkt('POINT(%s %s)' %(x,y)), winkel
